@@ -1,7 +1,6 @@
 
 import os, stat, time, weakref
 from zope.interface import implements
-from twisted.application import service
 from twisted.internet import defer
 from foolscap.api import Referenceable, DeadReferenceError, eventually
 import allmydata # for __full_version__
@@ -10,7 +9,7 @@ from allmydata.storage.server import si_b2a
 from allmydata.immutable import upload
 from allmydata.immutable.layout import ReadBucketProxy
 from allmydata.util.assertutil import precondition
-from allmydata.util import idlib, log, observer, fileutil, hashutil, dictutil
+from allmydata.util import log, observer, fileutil, hashutil, dictutil
 
 
 class NotEnoughWritersError(Exception):
@@ -54,23 +53,23 @@ class CHKCheckerAndUEBFetcher:
 
     def _get_all_shareholders(self, storage_index):
         dl = []
-        for (peerid, ss) in self._peer_getter(storage_index):
-            d = ss.callRemote("get_buckets", storage_index)
+        for s in self._peer_getter(storage_index):
+            d = s.get_rref().callRemote("get_buckets", storage_index)
             d.addCallbacks(self._got_response, self._got_error,
-                           callbackArgs=(peerid,))
+                           callbackArgs=(s,))
             dl.append(d)
         return defer.DeferredList(dl)
 
-    def _got_response(self, buckets, peerid):
+    def _got_response(self, buckets, server):
         # buckets is a dict: maps shum to an rref of the server who holds it
         shnums_s = ",".join([str(shnum) for shnum in buckets])
         self.log("got_response: [%s] has %d shares (%s)" %
-                 (idlib.shortnodeid_b2a(peerid), len(buckets), shnums_s),
+                 (server.get_name(), len(buckets), shnums_s),
                  level=log.NOISY)
         self._found_shares.update(buckets.keys())
         for k in buckets:
-            self._sharemap.add(k, peerid)
-        self._readers.update( [ (bucket, peerid)
+            self._sharemap.add(k, server.get_serverid())
+        self._readers.update( [ (bucket, server)
                                 for bucket in buckets.values() ] )
 
     def _got_error(self, f):
@@ -85,8 +84,8 @@ class CHKCheckerAndUEBFetcher:
         if not self._readers:
             self.log("no readers, so no UEB", level=log.NOISY)
             return
-        b,peerid = self._readers.pop()
-        rbp = ReadBucketProxy(b, peerid, si_b2a(self._storage_index))
+        b,server = self._readers.pop()
+        rbp = ReadBucketProxy(b, server, si_b2a(self._storage_index))
         d = rbp.get_uri_extension()
         d.addCallback(self._got_uri_extension)
         d.addErrback(self._ueb_error)
@@ -135,7 +134,8 @@ class CHKUploadHelper(Referenceable, upload.CHKUploader):
                 "application-version": str(allmydata.__full_version__),
                 }
 
-    def __init__(self, storage_index, helper,
+    def __init__(self, storage_index,
+                 helper, storage_broker, secret_holder,
                  incoming_file, encoding_file,
                  results, log_number):
         self._storage_index = storage_index
@@ -153,7 +153,8 @@ class CHKUploadHelper(Referenceable, upload.CHKUploader):
         self._helper.log("CHKUploadHelper starting for SI %s" % self._upload_id,
                          parent=log_number)
 
-        self._client = helper.parent
+        self._storage_broker = storage_broker
+        self._secret_holder = secret_holder
         self._fetcher = CHKCiphertextFetcher(self, incoming_file, encoding_file,
                                              self._log_number)
         self._reader = LocalCiphertextReader(self, storage_index, encoding_file)
@@ -477,7 +478,7 @@ class LocalCiphertextReader(AskUntilSuccessMixin):
 
 
 
-class Helper(Referenceable, service.MultiService):
+class Helper(Referenceable):
     implements(interfaces.RIHelper, interfaces.IStatsProducer)
     # this is the non-distributed version. When we need to have multiple
     # helpers, this object will become the HelperCoordinator, and will query
@@ -493,16 +494,17 @@ class Helper(Referenceable, service.MultiService):
     chk_upload_helper_class = CHKUploadHelper
     MAX_UPLOAD_STATUSES = 10
 
-    def __init__(self, basedir, stats_provider=None):
+    def __init__(self, basedir, storage_broker, secret_holder,
+                 stats_provider, history):
         self._basedir = basedir
+        self._storage_broker = storage_broker
+        self._secret_holder = secret_holder
         self._chk_incoming = os.path.join(basedir, "CHK_incoming")
         self._chk_encoding = os.path.join(basedir, "CHK_encoding")
         fileutil.make_dirs(self._chk_incoming)
         fileutil.make_dirs(self._chk_encoding)
         self._active_uploads = {}
         self._all_uploads = weakref.WeakKeyDictionary() # for debugging
-        self._all_upload_statuses = weakref.WeakKeyDictionary()
-        self._recent_upload_statuses = []
         self.stats_provider = stats_provider
         if stats_provider:
             stats_provider.register_producer(self)
@@ -513,15 +515,12 @@ class Helper(Referenceable, service.MultiService):
                           "chk_upload_helper.fetched_bytes": 0,
                           "chk_upload_helper.encoded_bytes": 0,
                           }
-        service.MultiService.__init__(self)
-
-    def setServiceParent(self, parent):
-        service.MultiService.setServiceParent(self, parent)
+        self._history = history
 
     def log(self, *args, **kwargs):
         if 'facility' not in kwargs:
             kwargs['facility'] = "tahoe.helper"
-        return self.parent.log(*args, **kwargs)
+        return log.msg(*args, **kwargs)
 
     def count(self, key, value=1):
         if self.stats_provider:
@@ -601,6 +600,8 @@ class Helper(Referenceable, service.MultiService):
             else:
                 self.log("creating new upload helper", parent=lp)
                 uh = self.chk_upload_helper_class(storage_index, self,
+                                                  self._storage_broker,
+                                                  self._secret_holder,
                                                   incoming_file, encoding_file,
                                                   r, lp)
                 self._active_uploads[storage_index] = uh
@@ -618,8 +619,8 @@ class Helper(Referenceable, service.MultiService):
         # see if this file is already in the grid
         lp2 = self.log("doing a quick check+UEBfetch",
                        parent=lp, level=log.NOISY)
-        sb = self.parent.get_storage_broker()
-        c = CHKCheckerAndUEBFetcher(sb.get_servers_for_index, storage_index, lp2)
+        sb = self._storage_broker
+        c = CHKCheckerAndUEBFetcher(sb.get_servers_for_psi, storage_index, lp2)
         d = c.check()
         def _checked(res):
             if res:
@@ -637,11 +638,9 @@ class Helper(Referenceable, service.MultiService):
 
     def _add_upload(self, uh):
         self._all_uploads[uh] = None
-        s = uh.get_upload_status()
-        self._all_upload_statuses[s] = None
-        self._recent_upload_statuses.append(s)
-        while len(self._recent_upload_statuses) > self.MAX_UPLOAD_STATUSES:
-            self._recent_upload_statuses.pop(0)
+        if self._history:
+            s = uh.get_upload_status()
+            self._history.notify_helper_upload(s)
 
     def upload_finished(self, storage_index, size):
         # this is called with size=0 if the upload failed
@@ -650,6 +649,3 @@ class Helper(Referenceable, service.MultiService):
         del self._active_uploads[storage_index]
         s = uh.get_upload_status()
         s.set_active(False)
-
-    def get_all_upload_statuses(self):
-        return self._all_upload_statuses
