@@ -1,10 +1,9 @@
-# -*- coding: utf-8 -*-
-
 from twisted.trial import unittest
 
 from twisted.python import usage, runtime
-from twisted.internet import utils
-import os.path, re, sys
+from twisted.internet import threads
+
+import os.path, re, sys, subprocess
 from cStringIO import StringIO
 from allmydata.util import fileutil, pollmixin
 from allmydata.util.encodingutil import unicode_to_argv, unicode_to_output, get_filesystem_encoding
@@ -15,30 +14,38 @@ import allmydata
 
 timeout = 240
 
+def get_root_from_file(src):
+    srcdir = os.path.dirname(os.path.dirname(os.path.normcase(os.path.realpath(src))))
+
+    root = os.path.dirname(srcdir)
+    if os.path.basename(srcdir) == 'site-packages':
+        if re.search(r'python.+\..+', os.path.basename(root)):
+            root = os.path.dirname(root)
+        root = os.path.dirname(root)
+    elif os.path.basename(root) == 'src':
+        root = os.path.dirname(root)
+
+    return root
+
 srcfile = allmydata.__file__
-srcdir = os.path.dirname(os.path.dirname(os.path.normcase(os.path.realpath(srcfile))))
+rootdir = get_root_from_file(srcfile)
 
-rootdir = os.path.dirname(srcdir)
-if os.path.basename(srcdir) == 'site-packages':
-    if re.search(r'python.+\..+', os.path.basename(rootdir)):
-        rootdir = os.path.dirname(rootdir)
-    rootdir = os.path.dirname(rootdir)
-elif os.path.basename(rootdir) == 'src':
-    rootdir = os.path.dirname(rootdir)
-
-bintahoe = os.path.join(rootdir, 'bin', 'tahoe')
-if sys.platform == "win32":
-    bintahoe += ".pyscript"
-    if not os.path.exists(bintahoe):
-       alt_bintahoe = os.path.join(rootdir, 'Scripts', 'tahoe.pyscript')
-       if os.path.exists(alt_bintahoe):
-           bintahoe = alt_bintahoe
+if hasattr(sys, 'frozen'):
+    bintahoe = os.path.join(rootdir, 'tahoe')
+    if sys.platform == "win32" and os.path.exists(bintahoe + '.exe'):
+        bintahoe += '.exe'
+else:
+    bintahoe = os.path.join(rootdir, 'bin', 'tahoe')
+    if sys.platform == "win32":
+        bintahoe += '.pyscript'
+        if not os.path.exists(bintahoe):
+            alt_bintahoe = os.path.join(rootdir, 'Scripts', 'tahoe.pyscript')
+            if os.path.exists(alt_bintahoe):
+                bintahoe = alt_bintahoe
 
 
-class SkipMixin:
+class RunBinTahoeMixin:
     def skip_if_cannot_run_bintahoe(self):
-        if "cygwin" in sys.platform.lower():
-            raise unittest.SkipTest("We don't know how to make this test work on cygwin: spawnProcess seems to hang forever. We don't know if 'bin/tahoe start' can be run on cygwin.")
         if not os.path.exists(bintahoe):
             raise unittest.SkipTest("The bin/tahoe script isn't to be found in the expected location (%s), and I don't want to test a 'tahoe' executable that I find somewhere else, in case it isn't the right executable for this version of Tahoe. Perhaps running 'setup.py build' again will help." % (bintahoe,))
 
@@ -48,18 +55,43 @@ class SkipMixin:
             # twistd on windows doesn't daemonize. cygwin should work normally.
             raise unittest.SkipTest("twistd does not fork under windows")
 
+    def run_bintahoe(self, args, stdin=None, python_options=[], env=None):
+        self.skip_if_cannot_run_bintahoe()
+
+        if hasattr(sys, 'frozen'):
+            if python_options:
+                raise unittest.SkipTest("This test doesn't apply to frozen builds.")
+            command = [bintahoe] + args
+        else:
+            command = [sys.executable] + python_options + [bintahoe] + args
+
+        if stdin is None:
+            stdin_stream = None
+        else:
+            stdin_stream = subprocess.PIPE
+
+        def _run():
+            p = subprocess.Popen(command, stdin=stdin_stream, stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=env)
+            (out, err) = p.communicate(stdin)
+            return (out, err, p.returncode)
+        return threads.deferToThread(_run)
+
+
+class BinTahoe(common_util.SignalMixin, unittest.TestCase, RunBinTahoeMixin):
+    def _check_right_code(self, file_to_check):
+        root_to_check = get_root_from_file(file_to_check)
+        if os.path.basename(root_to_check) == 'dist':
+            root_to_check = os.path.dirname(root_to_check)
 
-class BinTahoe(common_util.SignalMixin, unittest.TestCase, SkipMixin):
-    def test_the_right_code(self):
         cwd = os.path.normcase(os.path.realpath("."))
         root_from_cwd = os.path.dirname(cwd)
         if os.path.basename(root_from_cwd) == 'src':
             root_from_cwd = os.path.dirname(root_from_cwd)
 
-        same = (root_from_cwd == rootdir)
+        same = (root_from_cwd == root_to_check)
         if not same:
             try:
-                same = os.path.samefile(root_from_cwd, rootdir)
+                same = os.path.samefile(root_from_cwd, root_to_check)
             except AttributeError, e:
                 e  # hush pyflakes
 
@@ -67,9 +99,9 @@ class BinTahoe(common_util.SignalMixin, unittest.TestCase, SkipMixin):
             msg = ("We seem to be testing the code at %r,\n"
                    "(according to the source filename %r),\n"
                    "but expected to be testing the code at %r.\n"
-                   % (rootdir, srcfile, root_from_cwd))
+                   % (root_to_check, file_to_check, root_from_cwd))
 
-            root_from_cwdu = os.path.normcase(os.path.normpath(os.getcwdu()))
+            root_from_cwdu = os.path.dirname(os.path.normcase(os.path.normpath(os.getcwdu())))
             if os.path.basename(root_from_cwdu) == u'src':
                 root_from_cwdu = os.path.dirname(root_from_cwdu)
 
@@ -82,41 +114,53 @@ class BinTahoe(common_util.SignalMixin, unittest.TestCase, SkipMixin):
                 msg += "Please run the tests from the root of the Tahoe-LAFS distribution."
                 self.fail(msg)
 
-    def test_path(self):
-        self.skip_if_cannot_run_bintahoe()
-        d = utils.getProcessOutputAndValue(bintahoe, args=["--version-and-path"], env=os.environ)
+    def test_the_right_code(self):
+        self._check_right_code(srcfile)
+
+    def test_import_in_repl(self):
+        d = self.run_bintahoe(["debug", "repl"],
+                              stdin="import allmydata; print; print allmydata.__file__")
         def _cb(res):
+            out, err, rc_or_sig = res
+            self.failUnlessEqual(rc_or_sig, 0, str(res))
+            lines = out.splitlines()
+            self.failUnlessIn('>>>', lines[0], str(res))
+            self._check_right_code(lines[1])
+        d.addCallback(_cb)
+        return d
+    # The timeout was exceeded on FreeStorm's CentOS:
+    # http://tahoe-lafs.org/buildbot/builders/FreeStorm%20CentOS5-i386/builds/503/steps/test/logs/stdio
+    test_import_in_repl.timeout = 480
+
+    def test_path(self):
+        d = self.run_bintahoe(["--version-and-path"])
+        def _cb(res):
+            from allmydata import normalized_version
+
             out, err, rc_or_sig = res
             self.failUnlessEqual(rc_or_sig, 0, str(res))
 
             # Fail unless the allmydata-tahoe package is *this* version *and*
             # was loaded from *this* source directory.
 
-            verstr = str(allmydata.__version__)
+            required_verstr = str(allmydata.__version__)
 
-            self.failIfEqual(verstr, "unknown",
+            self.failIfEqual(required_verstr, "unknown",
                              "We don't know our version, because this distribution didn't come "
                              "with a _version.py and 'setup.py darcsver' hasn't been run.")
 
-            # The Python "rational version numbering" convention
-            # disallows "-r$REV" but allows ".post$REV"
-            # instead. Eventually we'll probably move to that. When we
-            # do, this test won't go red:
+            srcdir = os.path.dirname(os.path.dirname(os.path.normcase(os.path.realpath(srcfile))))
+            info = repr((res, allmydata.__appname__, required_verstr, srcdir))
 
-            ix = verstr.rfind('-r')
-            if ix != -1:
-                altverstr = verstr[:ix] + '.post' + verstr[ix+2:]
-            else:
-                ix = verstr.rfind('.post')
-                if ix != -1:
-                    altverstr = verstr[:ix] + '-r' + verstr[ix+5:]
-                else:
-                    altverstr = verstr
+            appverpath = out.split(')')[0]
+            (appver, path) = appverpath.split(' (')
+            (app, ver) = appver.split(': ')
 
-            required_ver_and_path = "%s: %s (%s)" % (allmydata.__appname__, verstr, srcdir)
-            alt_required_ver_and_path = "%s: %s (%s)" % (allmydata.__appname__, altverstr, srcdir)
-
-            self.failUnless(out.startswith(required_ver_and_path) or out.startswith(alt_required_ver_and_path), (out, err, rc_or_sig, required_ver_and_path))
+            self.failUnlessEqual(app, allmydata.__appname__, info)
+            norm_ver = normalized_version(ver)
+            norm_required = normalized_version(required_verstr)
+            self.failUnlessEqual(norm_ver, norm_required, info)
+            self.failUnlessEqual(path, srcdir, info)
         d.addCallback(_cb)
         return d
 
@@ -130,7 +174,7 @@ class BinTahoe(common_util.SignalMixin, unittest.TestCase, SkipMixin):
         except UnicodeEncodeError:
             raise unittest.SkipTest("A non-ASCII argument/output could not be encoded on this platform.")
 
-        d = utils.getProcessOutputAndValue(bintahoe, args=[tricky_arg], env=os.environ)
+        d = self.run_bintahoe([tricky_arg])
         def _cb(res):
             out, err, rc_or_sig = res
             self.failUnlessEqual(rc_or_sig, 1, str(res))
@@ -139,11 +183,8 @@ class BinTahoe(common_util.SignalMixin, unittest.TestCase, SkipMixin):
         return d
 
     def test_run_with_python_options(self):
-        self.skip_if_cannot_run_bintahoe()
-
         # -t is a harmless option that warns about tabs.
-        d = utils.getProcessOutputAndValue(sys.executable, args=['-t', bintahoe, '--version'],
-                                           env=os.environ)
+        d = self.run_bintahoe(["--version"], python_options=["-t"])
         def _cb(res):
             out, err, rc_or_sig = res
             self.failUnlessEqual(rc_or_sig, 0, str(res))
@@ -153,13 +194,14 @@ class BinTahoe(common_util.SignalMixin, unittest.TestCase, SkipMixin):
 
     def test_version_no_noise(self):
         self.skip_if_cannot_run_bintahoe()
-        import pkg_resources
-        try:
-            pkg_resources.require("Twisted>=9.0.0")
-        except pkg_resources.VersionConflict:
+
+        from allmydata import get_package_versions, normalized_version
+        twisted_ver = get_package_versions()['Twisted']
+
+        if not normalized_version(twisted_ver) >= normalized_version('9.0.0'):
             raise unittest.SkipTest("We pass this test only with Twisted >= v9.0.0")
 
-        d = utils.getProcessOutputAndValue(bintahoe, args=["--version"], env=os.environ)
+        d = self.run_bintahoe(["--version"])
         def _cb(res):
             out, err, rc_or_sig = res
             self.failUnlessEqual(rc_or_sig, 0, str(res))
@@ -209,11 +251,14 @@ class CreateNode(unittest.TestCase):
             # 'create-node', and disabled for 'create-client'.
             tahoe_cfg = os.path.join(n1, "tahoe.cfg")
             self.failUnless(os.path.exists(tahoe_cfg))
-            content = open(tahoe_cfg).read()
+            content = fileutil.read(tahoe_cfg).replace('\r\n', '\n')
             if kind == "client":
-                self.failUnless("\n[storage]\nenabled = false\n" in content)
+                self.failUnless(re.search(r"\n\[storage\]\n#.*\nenabled = false\n", content), content)
             else:
-                self.failUnless("\n[storage]\nenabled = true\n" in content)
+                self.failUnless(re.search(r"\n\[storage\]\n#.*\nenabled = true\n", content), content)
+                self.failUnless("\nreserved_space = 1G\n" in content)
+
+            self.failUnless(re.search(r"\n\[drop_upload\]\n#.*\nenabled = false\n", content), content)
 
         # creating the node a second time should be rejected
         rc, out, err = self.run_tahoe(argv)
@@ -285,7 +330,7 @@ class CreateNode(unittest.TestCase):
 
 
 class RunNode(common_util.SignalMixin, unittest.TestCase, pollmixin.PollMixin,
-              SkipMixin):
+              RunBinTahoeMixin):
     # exercise "tahoe start", for both introducer, client node, and
     # key-generator, by spawning "tahoe start" as a subprocess. This doesn't
     # get us figleaf-based line-level coverage, but it does a better job of
@@ -310,26 +355,36 @@ class RunNode(common_util.SignalMixin, unittest.TestCase, pollmixin.PollMixin,
         HOTLINE_FILE = os.path.join(c1, "suicide_prevention_hotline")
         TWISTD_PID_FILE = os.path.join(c1, "twistd.pid")
         INTRODUCER_FURL_FILE = os.path.join(c1, "introducer.furl")
+        PORTNUM_FILE = os.path.join(c1, "introducer.port")
+        NODE_URL_FILE = os.path.join(c1, "node.url")
+        CONFIG_FILE = os.path.join(c1, "tahoe.cfg")
 
-        d = utils.getProcessOutputAndValue(bintahoe, args=["--quiet", "create-introducer", "--basedir", c1], env=os.environ)
+        d = self.run_bintahoe(["--quiet", "create-introducer", "--basedir", c1])
         def _cb(res):
             out, err, rc_or_sig = res
             self.failUnlessEqual(rc_or_sig, 0)
+
+            # This makes sure that node.url is written, which allows us to
+            # detect when the introducer restarts in _node_has_restarted below.
+            config = fileutil.read(CONFIG_FILE)
+            self.failUnlessIn('\nweb.port = \n', config)
+            fileutil.write(CONFIG_FILE, config.replace('\nweb.port = \n', '\nweb.port = 0\n'))
+
             # by writing this file, we get ten seconds before the node will
             # exit. This insures that even if the test fails (and the 'stop'
             # command doesn't work), the client should still terminate.
-            open(HOTLINE_FILE, "w").write("")
+            fileutil.write(HOTLINE_FILE, "")
             # now it's safe to start the node
         d.addCallback(_cb)
 
         def _then_start_the_node(res):
-            return utils.getProcessOutputAndValue(bintahoe, args=["--quiet", "start", c1], env=os.environ)
+            return self.run_bintahoe(["--quiet", "start", c1])
         d.addCallback(_then_start_the_node)
 
         def _cb2(res):
             out, err, rc_or_sig = res
 
-            open(HOTLINE_FILE, "w").write("")
+            fileutil.write(HOTLINE_FILE, "")
             errstr = "rc=%d, OUT: '%s', ERR: '%s'" % (rc_or_sig, out, err)
             self.failUnlessEqual(rc_or_sig, 0, errstr)
             self.failUnlessEqual(out, "", errstr)
@@ -350,16 +405,24 @@ class RunNode(common_util.SignalMixin, unittest.TestCase, pollmixin.PollMixin,
         d.addCallback(lambda res: self.poll(_node_has_started))
 
         def _started(res):
-            open(HOTLINE_FILE, "w").write("")
+            # read the introducer.furl and introducer.port files so we can check that their
+            # contents don't change on restart
+            self.furl = fileutil.read(INTRODUCER_FURL_FILE)
+            self.failUnless(os.path.exists(PORTNUM_FILE))
+            self.portnum = fileutil.read(PORTNUM_FILE)
+
+            fileutil.write(HOTLINE_FILE, "")
             self.failUnless(os.path.exists(TWISTD_PID_FILE))
+            self.failUnless(os.path.exists(NODE_URL_FILE))
+
             # rm this so we can detect when the second incarnation is ready
-            os.unlink(INTRODUCER_FURL_FILE)
-            return utils.getProcessOutputAndValue(bintahoe, args=["--quiet", "restart", c1], env=os.environ)
+            os.unlink(NODE_URL_FILE)
+            return self.run_bintahoe(["--quiet", "restart", c1])
         d.addCallback(_started)
 
         def _then(res):
             out, err, rc_or_sig = res
-            open(HOTLINE_FILE, "w").write("")
+            fileutil.write(HOTLINE_FILE, "")
             errstr = "rc=%d, OUT: '%s', ERR: '%s'" % (rc_or_sig, out, err)
             self.failUnlessEqual(rc_or_sig, 0, errstr)
             self.failUnlessEqual(out, "", errstr)
@@ -367,22 +430,31 @@ class RunNode(common_util.SignalMixin, unittest.TestCase, pollmixin.PollMixin,
         d.addCallback(_then)
 
         # again, the second incarnation of the node might not be ready yet,
-        # so poll until it is
-        d.addCallback(lambda res: self.poll(_node_has_started))
+        # so poll until it is. This time INTRODUCER_FURL_FILE already
+        # exists, so we check for the existence of NODE_URL_FILE instead.
+        def _node_has_restarted():
+            return os.path.exists(NODE_URL_FILE) and os.path.exists(PORTNUM_FILE)
+        d.addCallback(lambda res: self.poll(_node_has_restarted))
+
+        def _check_same_furl_and_port(res):
+            self.failUnless(os.path.exists(INTRODUCER_FURL_FILE))
+            self.failUnlessEqual(self.furl, fileutil.read(INTRODUCER_FURL_FILE))
+            self.failUnlessEqual(self.portnum, fileutil.read(PORTNUM_FILE))
+        d.addCallback(_check_same_furl_and_port)
 
         # now we can kill it. TODO: On a slow machine, the node might kill
-        # itself before we get a chance too, especially if spawning the
+        # itself before we get a chance to, especially if spawning the
         # 'tahoe stop' command takes a while.
         def _stop(res):
-            open(HOTLINE_FILE, "w").write("")
+            fileutil.write(HOTLINE_FILE, "")
             self.failUnless(os.path.exists(TWISTD_PID_FILE))
 
-            return utils.getProcessOutputAndValue(bintahoe, args=["--quiet", "stop", c1], env=os.environ)
+            return self.run_bintahoe(["--quiet", "stop", c1])
         d.addCallback(_stop)
 
         def _after_stopping(res):
             out, err, rc_or_sig = res
-            open(HOTLINE_FILE, "w").write("")
+            fileutil.write(HOTLINE_FILE, "")
             # the parent has exited by now
             errstr = "rc=%d, OUT: '%s', ERR: '%s'" % (rc_or_sig, out, err)
             self.failUnlessEqual(rc_or_sig, 0, errstr)
@@ -393,49 +465,50 @@ class RunNode(common_util.SignalMixin, unittest.TestCase, pollmixin.PollMixin,
             # gone by now.
             self.failIf(os.path.exists(TWISTD_PID_FILE))
         d.addCallback(_after_stopping)
-
-        def _remove_hotline(res):
-            os.unlink(HOTLINE_FILE)
-            return res
-        d.addBoth(_remove_hotline)
+        d.addBoth(self._remove, HOTLINE_FILE)
         return d
-    test_introducer.timeout = 480 # This hit the 120-second timeout on "François Lenny-armv5tel", then it hit a 240-second timeout on our feisty2.5 buildslave: http://allmydata.org/buildbot/builders/feisty2.5/builds/2381/steps/test/logs/test.log
+    test_introducer.timeout = 960
+
+    # This test hit the 120-second timeout on "Francois Lenny-armv5tel", then it hit a 240-second timeout on our feisty2.5 buildslave: http://allmydata.org/buildbot/builders/feisty2.5/builds/2381/steps/test/logs/test.log
+    # Then it hit the 480 second timeout on Francois's machine: http://tahoe-lafs.org/buildbot/builders/FranXois%20lenny-armv5tel/builds/449/steps/test/logs/stdio
 
     def test_client_no_noise(self):
         self.skip_if_cannot_daemonize()
-        import pkg_resources
-        try:
-            pkg_resources.require("Twisted>=9.0.0")
-        except pkg_resources.VersionConflict:
+
+        from allmydata import get_package_versions, normalized_version
+        twisted_ver = get_package_versions()['Twisted']
+
+        if not normalized_version(twisted_ver) >= normalized_version('9.0.0'):
             raise unittest.SkipTest("We pass this test only with Twisted >= v9.0.0")
+
         basedir = self.workdir("test_client_no_noise")
         c1 = os.path.join(basedir, "c1")
         HOTLINE_FILE = os.path.join(c1, "suicide_prevention_hotline")
         TWISTD_PID_FILE = os.path.join(c1, "twistd.pid")
-        PORTNUMFILE = os.path.join(c1, "client.port")
+        PORTNUM_FILE = os.path.join(c1, "client.port")
 
-        d = utils.getProcessOutputAndValue(bintahoe, args=["--quiet", "create-client", "--basedir", c1, "--webport", "0"], env=os.environ)
+        d = self.run_bintahoe(["--quiet", "create-client", "--basedir", c1, "--webport", "0"])
         def _cb(res):
             out, err, rc_or_sig = res
             errstr = "cc=%d, OUT: '%s', ERR: '%s'" % (rc_or_sig, out, err)
             assert rc_or_sig == 0, errstr
             self.failUnlessEqual(rc_or_sig, 0)
-            # By writing this file, we get forty seconds before the client will exit. This insures
+
+            # By writing this file, we get two minutes before the client will exit. This ensures
             # that even if the 'stop' command doesn't work (and the test fails), the client should
             # still terminate.
-            open(HOTLINE_FILE, "w").write("")
-            open(os.path.join(c1, "introducer.furl"), "w").write("pb://xrndsskn2zuuian5ltnxrte7lnuqdrkz@127.0.0.1:55617/introducer\n")
+            fileutil.write(HOTLINE_FILE, "")
             # now it's safe to start the node
         d.addCallback(_cb)
 
         def _start(res):
-            return utils.getProcessOutputAndValue(bintahoe, args=["--quiet", "start", c1], env=os.environ)
+            return self.run_bintahoe(["--quiet", "start", c1])
         d.addCallback(_start)
 
         def _cb2(res):
             out, err, rc_or_sig = res
             errstr = "cc=%d, OUT: '%s', ERR: '%s'" % (rc_or_sig, out, err)
-            open(HOTLINE_FILE, "w").write("")
+            fileutil.write(HOTLINE_FILE, "")
             self.failUnlessEqual(rc_or_sig, 0, errstr)
             self.failUnlessEqual(out, "", errstr) # If you emit noise, you fail this test.
             errlines = err.split("\n")
@@ -455,7 +528,7 @@ class RunNode(common_util.SignalMixin, unittest.TestCase, pollmixin.PollMixin,
         d.addCallback(_cb2)
 
         def _node_has_started():
-            return os.path.exists(PORTNUMFILE)
+            return os.path.exists(PORTNUM_FILE)
         d.addCallback(lambda res: self.poll(_node_has_started))
 
         # now we can kill it. TODO: On a slow machine, the node might kill
@@ -463,8 +536,9 @@ class RunNode(common_util.SignalMixin, unittest.TestCase, pollmixin.PollMixin,
         # 'tahoe stop' command takes a while.
         def _stop(res):
             self.failUnless(os.path.exists(TWISTD_PID_FILE), (TWISTD_PID_FILE, os.listdir(os.path.dirname(TWISTD_PID_FILE))))
-            return utils.getProcessOutputAndValue(bintahoe, args=["--quiet", "stop", c1], env=os.environ)
+            return self.run_bintahoe(["--quiet", "stop", c1])
         d.addCallback(_stop)
+        d.addBoth(self._remove, HOTLINE_FILE)
         return d
 
     def test_client(self):
@@ -473,27 +547,33 @@ class RunNode(common_util.SignalMixin, unittest.TestCase, pollmixin.PollMixin,
         c1 = os.path.join(basedir, "c1")
         HOTLINE_FILE = os.path.join(c1, "suicide_prevention_hotline")
         TWISTD_PID_FILE = os.path.join(c1, "twistd.pid")
-        PORTNUMFILE = os.path.join(c1, "client.port")
+        PORTNUM_FILE = os.path.join(c1, "client.port")
+        NODE_URL_FILE = os.path.join(c1, "node.url")
+        CONFIG_FILE = os.path.join(c1, "tahoe.cfg")
 
-        d = utils.getProcessOutputAndValue(bintahoe, args=["--quiet", "create-node", "--basedir", c1, "--webport", "0"], env=os.environ)
+        d = self.run_bintahoe(["--quiet", "create-node", "--basedir", c1, "--webport", "0"])
         def _cb(res):
             out, err, rc_or_sig = res
             self.failUnlessEqual(rc_or_sig, 0)
-            # By writing this file, we get sixty seconds before the client will exit. This insures
+
+            # Check that the --webport option worked.
+            config = fileutil.read(CONFIG_FILE)
+            self.failUnlessIn('\nweb.port = 0\n', config)
+
+            # By writing this file, we get two minutes before the client will exit. This ensures
             # that even if the 'stop' command doesn't work (and the test fails), the client should
             # still terminate.
-            open(HOTLINE_FILE, "w").write("")
-            open(os.path.join(c1, "introducer.furl"), "w").write("pb://xrndsskn2zuuian5ltnxrte7lnuqdrkz@127.0.0.1:55617/introducer\n")
+            fileutil.write(HOTLINE_FILE, "")
             # now it's safe to start the node
         d.addCallback(_cb)
 
         def _start(res):
-            return utils.getProcessOutputAndValue(bintahoe, args=["--quiet", "start", c1], env=os.environ)
+            return self.run_bintahoe(["--quiet", "start", c1])
         d.addCallback(_start)
 
         def _cb2(res):
             out, err, rc_or_sig = res
-            open(HOTLINE_FILE, "w").write("")
+            fileutil.write(HOTLINE_FILE, "")
             errstr = "rc=%d, OUT: '%s', ERR: '%s'" % (rc_or_sig, out, err)
             self.failUnlessEqual(rc_or_sig, 0, errstr)
             self.failUnlessEqual(out, "", errstr)
@@ -510,22 +590,26 @@ class RunNode(common_util.SignalMixin, unittest.TestCase, pollmixin.PollMixin,
         d.addCallback(_cb2)
 
         def _node_has_started():
-            return os.path.exists(PORTNUMFILE)
+            return os.path.exists(NODE_URL_FILE) and os.path.exists(PORTNUM_FILE)
         d.addCallback(lambda res: self.poll(_node_has_started))
 
         def _started(res):
-            open(HOTLINE_FILE, "w").write("")
-            self.failUnless(os.path.exists(TWISTD_PID_FILE))
-            # rm this so we can detect when the second incarnation is ready
-            os.unlink(PORTNUMFILE)
+            # read the client.port file so we can check that its contents
+            # don't change on restart
+            self.portnum = fileutil.read(PORTNUM_FILE)
 
-            return utils.getProcessOutputAndValue(bintahoe, args=["--quiet", "restart", c1], env=os.environ)
+            fileutil.write(HOTLINE_FILE, "")
+            self.failUnless(os.path.exists(TWISTD_PID_FILE))
+
+            # rm this so we can detect when the second incarnation is ready
+            os.unlink(NODE_URL_FILE)
+            return self.run_bintahoe(["--quiet", "restart", c1])
         d.addCallback(_started)
 
         def _cb3(res):
             out, err, rc_or_sig = res
 
-            open(HOTLINE_FILE, "w").write("")
+            fileutil.write(HOTLINE_FILE, "")
             errstr = "rc=%d, OUT: '%s', ERR: '%s'" % (rc_or_sig, out, err)
             self.failUnlessEqual(rc_or_sig, 0, errstr)
             self.failUnlessEqual(out, "", errstr)
@@ -536,19 +620,23 @@ class RunNode(common_util.SignalMixin, unittest.TestCase, pollmixin.PollMixin,
         # so poll until it is
         d.addCallback(lambda res: self.poll(_node_has_started))
 
+        def _check_same_port(res):
+            self.failUnlessEqual(self.portnum, fileutil.read(PORTNUM_FILE))
+        d.addCallback(_check_same_port)
+
         # now we can kill it. TODO: On a slow machine, the node might kill
-        # itself before we get a chance too, especially if spawning the
+        # itself before we get a chance to, especially if spawning the
         # 'tahoe stop' command takes a while.
         def _stop(res):
-            open(HOTLINE_FILE, "w").write("")
+            fileutil.write(HOTLINE_FILE, "")
             self.failUnless(os.path.exists(TWISTD_PID_FILE), (TWISTD_PID_FILE, os.listdir(os.path.dirname(TWISTD_PID_FILE))))
-            return utils.getProcessOutputAndValue(bintahoe, args=["--quiet", "stop", c1], env=os.environ)
+            return self.run_bintahoe(["--quiet", "stop", c1])
         d.addCallback(_stop)
 
         def _cb4(res):
             out, err, rc_or_sig = res
 
-            open(HOTLINE_FILE, "w").write("")
+            fileutil.write(HOTLINE_FILE, "")
             # the parent has exited by now
             errstr = "rc=%d, OUT: '%s', ERR: '%s'" % (rc_or_sig, out, err)
             self.failUnlessEqual(rc_or_sig, 0, errstr)
@@ -559,18 +647,19 @@ class RunNode(common_util.SignalMixin, unittest.TestCase, pollmixin.PollMixin,
             # gone by now.
             self.failIf(os.path.exists(TWISTD_PID_FILE))
         d.addCallback(_cb4)
-        def _remove_hotline(res):
-            os.unlink(HOTLINE_FILE)
-            return res
-        d.addBoth(_remove_hotline)
+        d.addBoth(self._remove, HOTLINE_FILE)
         return d
 
+    def _remove(self, res, file):
+        fileutil.remove(file)
+        return res
+
     def test_baddir(self):
         self.skip_if_cannot_daemonize()
         basedir = self.workdir("test_baddir")
         fileutil.make_dirs(basedir)
 
-        d = utils.getProcessOutputAndValue(bintahoe, args=["--quiet", "start", "--basedir", basedir], env=os.environ)
+        d = self.run_bintahoe(["--quiet", "start", "--basedir", basedir])
         def _cb(res):
             out, err, rc_or_sig = res
             self.failUnlessEqual(rc_or_sig, 1)
@@ -578,7 +667,7 @@ class RunNode(common_util.SignalMixin, unittest.TestCase, pollmixin.PollMixin,
         d.addCallback(_cb)
 
         def _then_stop_it(res):
-            return utils.getProcessOutputAndValue(bintahoe, args=["--quiet", "stop", "--basedir", basedir], env=os.environ)
+            return self.run_bintahoe(["--quiet", "stop", "--basedir", basedir])
         d.addCallback(_then_stop_it)
 
         def _cb2(res):
@@ -589,7 +678,7 @@ class RunNode(common_util.SignalMixin, unittest.TestCase, pollmixin.PollMixin,
 
         def _then_start_in_bogus_basedir(res):
             not_a_dir = os.path.join(basedir, "bogus")
-            return utils.getProcessOutputAndValue(bintahoe, args=["--quiet", "start", "--basedir", not_a_dir], env=os.environ)
+            return self.run_bintahoe(["--quiet", "start", "--basedir", not_a_dir])
         d.addCallback(_then_start_in_bogus_basedir)
 
         def _cb3(res):
@@ -606,14 +695,14 @@ class RunNode(common_util.SignalMixin, unittest.TestCase, pollmixin.PollMixin,
         TWISTD_PID_FILE = os.path.join(c1, "twistd.pid")
         KEYGEN_FURL_FILE = os.path.join(c1, "key_generator.furl")
 
-        d = utils.getProcessOutputAndValue(bintahoe, args=["--quiet", "create-key-generator", "--basedir", c1], env=os.environ)
+        d = self.run_bintahoe(["--quiet", "create-key-generator", "--basedir", c1])
         def _cb(res):
             out, err, rc_or_sig = res
             self.failUnlessEqual(rc_or_sig, 0)
         d.addCallback(_cb)
 
         def _start(res):
-            return utils.getProcessOutputAndValue(bintahoe, args=["--quiet", "start", c1], env=os.environ)
+            return self.run_bintahoe(["--quiet", "start", c1])
         d.addCallback(_start)
 
         def _cb2(res):
@@ -641,7 +730,7 @@ class RunNode(common_util.SignalMixin, unittest.TestCase, pollmixin.PollMixin,
             self.failUnless(os.path.exists(TWISTD_PID_FILE))
             # rm this so we can detect when the second incarnation is ready
             os.unlink(KEYGEN_FURL_FILE)
-            return utils.getProcessOutputAndValue(bintahoe, args=["--quiet", "restart", c1], env=os.environ)
+            return self.run_bintahoe(["--quiet", "restart", c1])
         d.addCallback(_started)
 
         def _cb3(res):
@@ -661,7 +750,7 @@ class RunNode(common_util.SignalMixin, unittest.TestCase, pollmixin.PollMixin,
         # 'tahoe stop' command takes a while.
         def _stop(res):
             self.failUnless(os.path.exists(TWISTD_PID_FILE))
-            return utils.getProcessOutputAndValue(bintahoe, args=["--quiet", "stop", c1], env=os.environ)
+            return self.run_bintahoe(["--quiet", "stop", c1])
         d.addCallback(_stop)
 
         def _cb4(res):
